@@ -90,7 +90,7 @@ loggers:
         level: DEBUG
         propogate: true
     "":
-        level: DEBUG
+        level: INFO
         handlers:
             - console
 """
@@ -124,9 +124,10 @@ def _check_call(params, cwd, log, *args, **kw):
         output = subprocess.check_output(
             params, cwd=cwd, stderr=subprocess.STDOUT, env=os.environ)
     except subprocess.CalledProcessError, e:
-        print " ".join(params), cwd, " ".join(
-            ["%s=%s" % (k, os.environ[k]) for k in os.environ if k.startswith("JUJU")])
-        print e.output
+        #print "subprocess error"
+        #print " ".join(params), "\ncwd: %s\n" % cwd, "\n ".join(
+        #    ["%s=%s" % (k, os.environ[k]) for k in os.environ if k.startswith("JUJU")])
+        #print e.output
         log.error(*args, **kw)
         log.error("Command (%s) Output:\n\n %s", " ".join(params), e.output)
         raise ErrorExit(e)
@@ -193,12 +194,12 @@ def setup_parser():
     parser.add_option('-D', '--destroy-services',
                       help='Destroy all services (do not terminate machines)',
                       dest="destroy_services", action="store_true",
-                      default="False")
+                      default=False)
     parser.add_option('-T', '--terminate-machines',
                       help=('Terminate all machines but the bootstrap node.  '
                             'Destroy any services that exist on each'),
                       dest="terminate_machines", action="store_true",
-                      default="False")
+                      default=False)
     parser.add_option('-t', '--timeout',
                       help='Timeout (sec) for entire deployment (45min default)',
                       dest='timeout', action='store', type='int', default=2700)
@@ -225,11 +226,13 @@ def setup_parser():
     parser.add_option('-W', '--watch', help='Watch environment changes on console',
                       dest="watch",
                       action="store_true", default=False)
-    parser.add_option('-r', '--revdate', help='Deploy charm revisions as of date',
-                      dest="revspec",
-                      action="store", default=False)
+
+# bzr up -r doesn't seem to support revisionspec date forms??
+#    parser.add_option('-r', '--revdate', help='Deploy charm revisions as of date',
+#                      dest="revspec",
+#                      action="store", default=False)
     parser.add_option('-w', '--relation-wait', action='store', dest='rel_wait',
-                      default='60',
+                      default=60, type=int,
                       help=('Number of seconds to wait before checking for '
                             'relation errors after all relations have been added '
                             'and subordinates started. (default: 60)'))
@@ -336,6 +339,7 @@ class Vcs(object):
             output = subprocess.check_output(
                 args, cwd=cwd or self.path, stderr=subprocess.STDOUT)
         except subprocess.CalledProcessError, e:
+            print "vcs err", " ".join(args), "[dir: %s]" % cwd
             self.log.error(error_msg % self.get_err_msg_ctx(e))
             raise ErrorExit()
         return output.strip()
@@ -426,9 +430,6 @@ class Charm(object):
                 name, store_url, branch)
         return cls(name, charm_path, branch, rev, build, store_url)
 
-    def set_revision(self, revspec):
-        self.rev = revspec
-
     def exists(self):
         return path_exists(self.path)
 
@@ -451,7 +452,7 @@ class Charm(object):
     def fetch(self):
         if not self.branch:
             return
-        self.log.debug("Branching charm  %s @ %s", self.branch, self.path)
+        self.log.debug(" Branching charm %s @ %s", self.branch, self.path)
         self.vcs.branch()
         self.build()
 
@@ -459,12 +460,12 @@ class Charm(object):
         if not self.branch:
             return
         assert self.exists()
-        self.log.debug("Updating charm %s from %s", self.path, self.branch)
+        self.log.debug(" Updating charm %s from %s", self.path, self.branch)
         self.vcs.update(self.rev)
         if build:
             self.build()
 
-    def has_local_mods(self):
+    def is_modified(self):
         if not self.branch:
             return False
         return self.vcs.is_modified()
@@ -549,13 +550,13 @@ class Deployment(object):
         svc_data = self.data['services'][svc_name]
         return Charm.from_service(svc_name, self.series_path, svc_data)
 
-    def fetch_charms(self, update=False, no_local_mods=False, revspec=None):
+    def fetch_charms(self, update=False, no_local_mods=False):
+        if not os.path.exists(self.series_path):
+            os.mkdir(self.series_path)
         for charm in self.get_charms():
-            if revspec is not None:
-                charm.set_revision(revspec)
             if charm.exists():
                 if no_local_mods:
-                    if charm.has_local_changes():
+                    if charm.is_modified():
                         self.log.warning(
                             "Charm %r has local modifications",
                             charm.path)
@@ -638,7 +639,6 @@ class Environment(object):
             # should really have a cheaper/faster way of getting the endpoint
             # ala juju status 0 for a get_endpoint method.
             self.get_cli_status()
-
         while True:
             try:
                 self.client = EnvironmentClient(self.api_endpoint)
@@ -649,6 +649,7 @@ class Environment(object):
             else:
                 break
         self.client.login(self._get_token())
+        self.log.debug("Connected to environment")
 
     def _get_token(self):
         with open(self.env_config_path) as fh:
@@ -687,18 +688,16 @@ class Environment(object):
                 fh.write(yaml_dump(config))
                 params.extend(["--config", fh.name])
             if constraints:
-                params.extend(["--constraints", '"%s"' % constraints])
+                params.extend(['--constraints', constraints])
+                #print 'constraints', constraints
             if num_units != 1:
                 params.extend(["--num-units", str(num_units)])
             if charm_url.startswith('local'):
-                assert repo
-                params.extend(["--repository", repo])
+                if repo == "":
+                    repo = "."
+                params.extend(["--repository=%s" % repo])
             params.extend([charm_url, name])
-
-            ec = subprocess.call(params)
-            if ec != 0:
-                self.log.error("Error deploying service %r", name)
-                raise ErrorExit()
+            _check_call(params, '', self.log, "Error deploying service %r", name)
 
     def add_units(self, service_name, num_units):
         return self.client.add_units(service_name, num_units)
@@ -711,7 +710,8 @@ class Environment(object):
 
         The machine can't have any running units, after removing the units or
         destroying the service, use wait_for_units to know when its safe to
-        delete the machine.
+        delete the machine (ie units have finished executing stop hooks and are
+        removed)
         """
         if int(mid) == 0:
             raise RuntimeError("Can't terminate machine 0")
@@ -722,8 +722,13 @@ class Environment(object):
     def get_service_address(self, svc_name):
         status = self.get_cli_status()
         if svc_name not in status['services']:
+            self.log.warning("Service %s does not exist", svc_name)
             return None
-        return status['services']['units'][0].get('public-address', '')
+        units = status['services'][svc_name].get('units', {})
+        unit_keys = list(sorted(units.keys()))
+        if unit_keys:
+            return units[unit_keys[0]].get('public-address', '')
+        self.log.warning("Service %s has no units")
 
     def get_cli_status(self):
         params = self._named_env(["juju", "status"])
@@ -731,27 +736,40 @@ class Environment(object):
         status = yaml_load(output)
         # Opportunistic, see connect method comment.
         if not self.api_endpoint:
-            self.api_endpoint = "wss://%s:17017/" % (
+            self.api_endpoint = "wss://%s:17070/" % (
                 status["machines"]["0"]["dns-name"])
         return status
 
-    def reset(self, terminate_machines=False, terminate_delay=0, timeout=60, watch=False):
-        self.log.info("Resetting environment")
-        status = self.get_cli_status()
-
-        for s in status['services'].keys():
+    def reset(self, terminate_machines=False, terminate_delay=0, timeout=360, watch=False):
+        status = self.status()
+        destroyed = False
+        for s in status.get('services', {}).keys():
+            self.log.debug(" Destroying service %s", s)
             self.client.destroy_service(s)
+            destroyed = True
+            # Check for any units in error and resolve them.
+            for uid, u in status['services'][s].get('units', {}).items():
+                if u['agent-state'] == 'error':
+                    self.log.debug("  Resolving unit error %s for destroy", uid)
+                    self.client.resolved(uid)
 
         # Wait for units
-        self.wait_for_units(timeout, "removed", watch=watch)
+        if destroyed:
+            self.wait_for_units(timeout, "removed", watch=watch)
+
+        # The only value to not terminating is keeping the data on the machines around.
+        if not terminate_machines:
+            self.log.info(" *juju-core machines are not reusable for units")
+            return
 
         # Terminate machines
+        self.log.info("Terminating machines")
         for mid in status['machines'].keys():
             if mid == "0":
                 continue
+            self.log.debug(" Terminating machine %s", mid)
             self.terminate_machine(mid)
             if terminate_delay:
-                self.log.debug("terminate delay %d", terminate_delay)
                 time.sleep(terminate_delay)
 
     def status(self):
@@ -766,11 +784,11 @@ class Environment(object):
             self.client.wait_for_units(timeout, goal_state, callback=callback)
         except UnitErrors, e:
             error_units = [
-                "unit: %s: machine: %s agent-state: %s/%s details: %s" % (
-                    u['Name'], u['Machine'], u['Status'], u['Life'], u['StatusInfo']
+                "unit: %s: machine: %s agent-state: %s details: %s" % (
+                    u['Name'], u['MachineId'], u['Status'], u['StatusInfo']
                 )
                 for u in e.errors]
-            self.log.errors("The following units had errors:\n   %s" % (
+            self.log.error("The following units had errors:\n   %s" % (
                 "   \n".join(error_units)))
             raise ErrorExit()
 
@@ -784,7 +802,7 @@ class Environment(object):
             if ct == "remove":
                 state = "removed"
         self.log.debug(
-            "Delta %s: %s %s:%s", et, name, ct, state)
+            " Delta %s: %s %s:%s", et, name, ct, state)
 
     def _format_endpoints(self, eps):
         if len(eps) == 1:
@@ -811,45 +829,45 @@ class Importer(object):
         self.deployment = deployment
 
     def add_units(self):
+        # TODO Check if subordinate
         self.log.debug("Adding units...")
         # Add units to existing services that don't match count.
         env_status = self.env.status()
         added = set()
         for svc in self.deployment.get_services():
-            delta = svc.num_units - len(env_status['service']['units'])
-        if delta < 0:
-            self.log.info(
-                "Adding %d more units to %s" % (abs(delta), svc.name))
-            for u in self.env.add_units(svc.name, abs(delta)):
-                added.add(u)
-        else:
-            self.log.debug(
-                "Service %r does not need any more units added." % svc.name)
+            delta = svc.num_units - len(env_status['services'][svc.name]['units'])
+            if delta > 0:
+                self.log.info(
+                    "Adding %d more units to %s" % (abs(delta), svc.name))
+                for u in self.env.add_units(svc.name, abs(delta)):
+                    added.add(u)
+            else:
+                self.log.debug(
+                    " Service %r does not need any more units added." % svc.name)
 
     def get_charms(self):
         # Get Charms
         self.log.debug("Getting charms...")
         self.deployment.fetch_charms(
             update=self.options.update_charms,
-            no_local_mods=self.options.no_local_mods,
-            revspec=self.options.revspec)
+            no_local_mods=self.options.no_local_mods)
         # Load config overrides after we can validate them.
-        self.deployment.load_overrides(self.options.overrides)
+        self.deployment.load_overrides(self.options.overrides or ())
         # Load includes
         self.deployment.resolve_config()
 
     def deploy_services(self):
-        self.log.debug("Deploying services...")
+        self.log.info("Deploying services...")
         env_status = self.env.status()
         for svc in self.deployment.get_services():
             if svc.name in env_status['services']:
-                self.log.info(
-                    "Service %r already deployed. Skipping" % svc)
+                self.log.debug(
+                    " Service %r already deployed. Skipping" % svc.name)
                 continue
 
             charm = self.deployment.get_charm_for(svc.name)
-            self.log.debug("Deploying service %s using %s")
-            self.env.deploy_service(
+            self.log.info(" Deploying service %s using %s", svc.name, charm.charm_url)
+            self.env.deploy(
                 svc.name,
                 charm.charm_url,
                 self.deployment.repo_path,
@@ -862,27 +880,63 @@ class Importer(object):
 
     def add_relations(self):
         self.log.info("Adding relations...")
-        # Relations
-        for end_a, end_b in self.deployment.get_relations():
-            self.log.debug(" Adding relation %s <-> %s", end_a, end_b)
-            self.env.add_relation(end_a, end_b)
-            time.sleep(5)  # per the original
 
-        self.log.debug("Waiting for relations %d", self.options.rel_wait)
-        time.sleep(self.options.rel_wait)
+        # Relations
+        status = self.env.status()
+        created = False
+
+        for end_a, end_b in self.deployment.get_relations():
+            if self._rel_exists(status, end_a, end_b):
+                continue
+            self.log.info(" Adding relation %s <-> %s", end_a, end_b)
+            self.env.add_relation(end_a, end_b)
+            created = True
+            time.sleep(5)  # per the original
+        return created
+
+    def _rel_exists(self, status, end_a, end_b):
+        # Checks for a named relation on one side that matches the local
+        # endpoint and remote service.
+        (name_a, name_b, rem_a, rem_b) = (end_a, end_b, None, None)
+
+        if ":" in end_a:
+            name_a, rem_a = end_a.split(":", 1)
+        if ":" in end_b:
+            name_b, rem_b = end_b.split(":", 1)
+
+        rels_svc_a = status['services'][name_a].get('relations', {})
+
+        found = False
+        for r, related in rels_svc_a.items():
+            if name_b in related:
+                if rem_a and not r in rem_a:
+                    continue
+                found = True
+                break
+        if found:
+            return True
+        return False
 
     def wait_for_units(self):
         timeout = self.options.timeout - (time.time() - self.start_time)
-        self.env.wait_for_units(timeout)
+        if timeout < 0:
+            self.log.error("Reached deployment timeout.. exiting")
+            raise ErrorExit()
+        self.env.wait_for_units(int(timeout), watch=self.options.watch)
 
     def run(self):
         self.start_time = time.time()
+        self.env.connect()
         self.get_charms()
         self.deploy_services()
         self.wait_for_units()
         self.add_units()
-        self.add_relations()
+        rels_created = self.add_relations()
+        # Wait for the units to be up before waiting for rel stability.
         self.wait_for_units()
+        if rels_created:
+            self.log.debug("Waiting for relations %d", self.options.rel_wait)
+            time.sleep(self.options.rel_wait)
 
 
 def main():
@@ -894,7 +948,7 @@ def main():
     start_time = time.time()
 
     env = Environment(options.juju_env)
-    config = ConfigStack(options.config_files)
+    config = ConfigStack(options.configs or [])
 
     # Debug implies watching
     if options.debug and not options.watch:
@@ -903,7 +957,10 @@ def main():
     # Destroy services and exit
     if options.destroy_services or options.terminate_machines:
         log.debug("Resetting environment...")
-        env.reset(options.terminate_machines, options.deploy_delay, options.watch)
+        env.connect()
+        env.reset(terminate_machines=options.terminate_machines,
+                  terminate_delay=1,
+                  watch=options.watch)
         log.info("Environment reset")
         sys.exit(0)
 
